/* File:      mat_pp.C
   Date:      9/20/01
   Author:    C. Paciorek
   Purpose:   Member functions for template for creating vector of pointers to type <T>
   Compile:   gpp
   Usage:     Creates storages structure via constructor;
   .read reads data in by row or by column from matrix style data file or ragged data file
   << is overloaded to output the matrix with one line per row
   Errors:    error messages are returned if the number of values,lines, or values in a line does not match that expected based on the original construction
   Requires:  mat_pp.h             
*/

// THIS FILE IS NOT CURRENTLY IN USE due to issues with linking when using templates; all code for the template class and its member functions is in mat_pp.h

#include <fstream.h>
#include <iostream.h>
#include <iomanip.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strstream.h>
#include <string>
#include <iostream>
#include <math.h>
#include <stddef.h>
#include <unistd.h>
#include <ctime>
// these includes can be removed once I have base.h up and running

//#include "base.h"
#include "mat_pp.h"

template<class T>
ostream& operator<<(ostream& stream,mat_pp<T>* this_mat_pp)
{
  // outputs matrix with one line per row 

  int i,j;
  for(i=0;i<this_mat_pp->ptr_len;i++)
    {
      for(j=0;j<this_mat_pp->vec_len;j++)
	{
	  stream << this_mat_pp->vals[i][j] << " "; 
	}
      stream << endl;
    }
  stream << endl;
  return(stream);
}

  

template<class T>
mat_pp<T>::mat_pp(int ptr_len_in,int vec_len_in) {
  int i;
  ptr_len=ptr_len_in;
  vec_len=vec_len_in;
  if(vec_len*ptr_len*16>SIZE){
    mat_pp::warn(0,0,TOO_BIG);
  }
  else{  
    vals=new (T*)[ptr_len];
    for(i=0;i<ptr_len;i++)
      {
	vals[i]=new T[vec_len];
      }
  }
}

template<class T>
mat_pp<T>::~mat_pp(void)
{
  int i;
  for(i=0;i<ptr_len;i++)
    {
      delete [] vals[i];  // delete each array
    }
  delete [] vals;  // delete array of pointers to arrays
}

template<class T>
T* mat_pp<T>::getvec(int index,int rowcol){
  T* tmp;
  int i;
  if(rowcol==ROW){
    if(index<ptr_len){
      tmp = new T[vec_len];
      for(i=0;i<vec_len;i++){
	tmp[i]=vals[index][i];
      }
    }
    else{
      warn(index,0,INDEX_TOO_LARGE);
    }
  }
  else{
    if(index<vec_len){
      tmp = new T[ptr_len];
      for(i=0;i<ptr_len;i++){
	tmp[i]=vals[i][index];
      }
    }
    else{
      warn(index,0,INDEX_TOO_LARGE);
    }
  }
  return(tmp);
}

template<class T>
void mat_pp<T>::read(string file_fn,int orient,int format)
{
  // reads in data as determined by orient and format

  char *linehold = new char[SIZE];
  int line,incr,nonint,i,j,cnt,len;
  int *row; int *col; int* rowlen; int* collen;
  T *tmpvals = new T[vec_len*ptr_len+1];
  int *rowlengths=new int[vec_len*ptr_len+1];

  mat_pp::file_fn=file_fn;
  ifstream input(file_fn.c_str());
  if (input.fail()) 
    {
      cerr << "Error opening file: " << file_fn << endl;
    }
  line=0;
  nonint=0;
  cnt=0;

  while(!input.eof())
    {
      input.getline(linehold,SIZE);
      istrstream buffer(linehold);
      incr=0;
      while(!buffer.fail())
	{
	  buffer >> tmpvals[cnt];
	  if(!buffer.fail()) {incr++; cnt++;}  // don't increment if nothing read
	}
      rowlengths[line]=incr;
      if(rowlengths[line]) {line++;}  // don't worry about rows with no data

      // check for non-numbers:
      len=strlen(linehold);
      for(i=0;i<len;i++)
	{
	  if(! (isdigit(linehold[i]) || isspace(linehold[i])) )
	    {
	      if(! (
		 (linehold[i]=='-' && (isdigit(linehold[i+1]) || linehold[i]=='.') && (i==0 || isspace(linehold[i-1]))) ||
		 (linehold[i]=='.' && (isdigit(linehold[i+1])) &&
		  (i==0 || isdigit(linehold[i-1]) || isspace(linehold[i-1]) || linehold[i-1]=='-'))
		 ) )
		{
	       
		  mat_pp::warn(line,0,NAN);
		}
	      if(linehold[i]=='.') {nonint=1;}
	    }
	}
      //
      
    }


  // give warnings as necessary
  if(nonint){
    mat_pp::warn(0,0,NON_INTEGER);  // warn if not integers; not sure how to check type of T so that irrelevant warnings not give
  }
  if(format==MATRIX) {
    if((orient==BY_ROW && line!=ptr_len) || (orient==BY_COL && line!=vec_len)) {
      mat_pp::warn(0,line,WRONG_NUMBER_LINES);
    }   // number of lines wrong
    for(i=0;i<line;i++) {
      if(rowlengths[i] && (orient==BY_ROW && rowlengths[i]!=vec_len) || (orient==BY_COL && rowlengths[i]!=ptr_len)) {
	mat_pp::warn(i+1,rowlengths[i],LINE_WRONG_LENGTH);
      }  // row lengths wrong
    }
  }
  if(format==RAGGED){
    if(cnt!=vec_len*ptr_len){
      mat_pp::warn(0,cnt,WRONG_NUMBER_VALUES);
    }  // total number of values wrong
  }

  // fill vals
  if(orient==BY_ROW){
    row=&i; col=&j; rowlen=&ptr_len; collen=&vec_len;
  }
  else{
    row=&j;col=&i; rowlen=&vec_len; collen=&ptr_len;
  }
  cnt=0; i=0;
  while(i<*rowlen){
    j=0;
    while(j<*collen){
      vals[*row][*col]=tmpvals[cnt];
      cnt++; j++;
    }
    i++;
  }

  delete [] linehold;
  delete [] tmpvals;
  delete [] rowlengths;
}

template<class T>
void mat_pp<T>::warn(int line,int value,int warn)
{
  if(warn == NAN){
    cerr << "Warning: there is a non-number in " << mat_pp::file_fn << ", line " << line <<".\n";
  }
  if(warn == NON_INTEGER){
    cerr << "Warning: data " << mat_pp::file_fn << " are not integer-valued.\n";
  }
  if(warn == WRONG_NUMBER_LINES){
    cerr << "Warning: found " << value << " lines in " << mat_pp::file_fn << "; this does not match the number expected.\n";
  }
  if(warn == LINE_WRONG_LENGTH){
    cerr << "Warning: line " << line << " has " << value << " values in " << mat_pp::file_fn << "; this does not match the number expected.\n";
  }
  if(warn == WRONG_NUMBER_VALUES){
    cerr << "Warning: there are " << value << " values in " << mat_pp::file_fn << "; this does not match the number expected.\n";
  }
  if(warn==TOO_BIG)
    {
      cerr << "Need to allocate more space in temporary storage in mat_pp read function\n";
    }
  if(warn==INDEX_TOO_LARGE)
    {
      cerr << "Index " << line << " exceeds number of rows or columns\n";
    }
}



